@@ -1,8 +1,12 @@
 use crate::*;
-use futures::never::Never;
+use futures::future::select_all;
+use never::Never;
 use std::time::Duration;
 use std::{future::Future, time::Instant};
-use tokio::{self, select, time};
+use tokio::{
+    self, select,
+    time::{sleep, sleep_until},
+};
 
 pub fn map<E, I, O, F, Fut>(source: E, mut f: F) -> Eventual<O>
 where
@@ -25,56 +29,147 @@ pub fn timer(interval: Duration) -> Eventual<Instant> {
     Eventual::spawn(move |mut writer| async move {
         loop {
             writer.write(Instant::now());
-            time::sleep(interval).await;
+            sleep(interval).await;
         }
     })
 }
 
-// TODO: Put this in a macro to de-duplicate A/B and support more arguments.
-pub fn join<A, Ar, B, Br>(a: Ar, b: Br) -> Eventual<(A, B)>
-where
-    A: Value,
-    B: Value,
-    Ar: IntoReader<Output = A>,
-    Br: IntoReader<Output = B>,
-{
-    let mut a = a.into_reader();
-    let mut b = b.into_reader();
+pub trait Joinable {
+    type Output;
+    fn join(self) -> Eventual<Self::Output>;
+}
 
-    Eventual::spawn(move |mut writer| async move {
-        let mut count = 0;
-        let mut ab = (None, None);
-
-        let mut ab = loop {
-            select! {
-                a_value = a.next() => {
-                    if ab.0.replace(a_value?).is_none() {
-                        count += 1;
-                    }
-                }
-                b_value = b.next() => {
-                    if ab.1.replace(b_value?).is_none() {
-                        count += 1;
-                    }
-                }
-            }
-            if count == 2 {
-                break (ab.0.unwrap(), ab.1.unwrap());
-            }
-        };
-        loop {
-            writer.write(ab.clone());
-
-            select! {
-                a_value = a.next() => {
-                    ab.0 = a_value?;
-                }
-                b_value = b.next() => {
-                    ab.1 = b_value?;
-                }
+macro_rules! impl_tuple {
+    ($len:expr, $($T:ident, $t:ident),*) => {
+        impl<T, $($T,)*> Selectable for ($($T,)*)
+            where
+            $($T: IntoReader<Output = T>,)*
+            // TODO: I don't understand why this bound is required.
+            // It is always the case that this bound is satisfied.
+            Vec<EventualReader<T>>: Selectable,
+        {
+            type Output = <Vec<EventualReader<T>> as Selectable>::Output;
+            fn select(self) -> Eventual<Self::Output> {
+                let ($($t),*) = self;
+                $(let $t = $t.into_reader();)*
+                #[allow(deprecated)]
+                vec![$($t),*].select()
             }
         }
-    })
+
+        impl<$($T,)*> Joinable for ($($T,)*)
+            where
+                $($T: IntoReader,)*
+        {
+            type Output = ($($T::Output),*);
+
+            #[allow(non_snake_case)]
+            fn join(self) -> Eventual<Self::Output> {
+                let ($($T),*) = self;
+                $(let mut $T = $T.into_reader();)*
+
+                Eventual::spawn(move |mut writer| async move {
+                    // In the first section we wait until all values are available
+                    let mut count:usize = 0;
+                    $(let mut $t = None;)*
+                    let ($(mut $t,)*) = loop {
+                        select! {
+                            $(
+                                next = $T.next() => {
+                                    if $t.replace(next?).is_none() {
+                                        count += 1;
+                                    }
+                                }
+                            )*
+                        }
+                        if count == 2 {
+                            break ($($t.unwrap()),*);
+                        }
+                    };
+                    // Once all values are available, start writing but continue
+                    // to update.
+                    loop {
+                        writer.write(($($t.clone(),)*));
+
+                        select! {
+                            $(
+                                next = $T.next() => {
+                                    $t = next?;
+                                }
+                            )*
+                        }
+                    }
+                })
+            }
+        }
+    };
+}
+
+macro_rules! impl_tuples {
+    ($len:expr, $A:ident, $a:ident) => { };
+    ($len:expr, $A:ident, $a:ident, $($T:ident, $t:ident),+) => {
+        impl_tuple!($len, $A, $a, $($T, $t),+);
+        impl_tuples!($len - 1, $($T, $t),+);
+    }
+}
+
+impl_tuples!(12, A, a, B, b, C, c, D, d, E, e, F, f, G, g, H, h, I, i, J, j, K, k, L, l);
+
+pub fn join<J>(joinable: J) -> Eventual<J::Output>
+where
+    J: Joinable,
+{
+    joinable.join()
+}
+
+pub trait Selectable {
+    type Output;
+    #[deprecated = "Not deterministic. This doesn't seem as harmful as filter, because it doesn't appear to miss updates."]
+    fn select(self) -> Eventual<Self::Output>;
+}
+
+#[deprecated = "Not deterministic. This doesn't seem as harmful as filter, because it doesn't appear to miss updates."]
+pub fn select<S>(selectable: S) -> Eventual<S::Output>
+where
+    S: Selectable,
+{
+    #[allow(deprecated)]
+    selectable.select()
+}
+
+impl<R> Selectable for Vec<R>
+where
+    R: IntoReader,
+{
+    type Output = R::Output;
+    fn select(self) -> Eventual<Self::Output> {
+        // TODO: With specialization we can avoid what is essentially an
+        // unnecessary clone when R is EventualReader
+        let mut readers: Vec<_> = self.into_iter().map(|v| v.into_reader()).collect();
+        Eventual::spawn(move |mut writer| async move {
+            loop {
+                if readers.len() == 0 {
+                    return Err(Closed);
+                }
+                let read_futs: Vec<_> = readers.iter_mut().map(|r| r.next()).collect();
+
+                let (output, index, remainder) = select_all(read_futs).await;
+
+                // Ideally, we would want to re-use this list, but in most
+                // cases we can't because it may have been shuffled.
+                drop(remainder);
+
+                match output {
+                    Ok(value) => {
+                        writer.write(value);
+                    }
+                    Err(Closed) => {
+                        readers.remove(index);
+                    }
+                }
+            }
+        })
+    }
 }
 
 pub fn throttle<E>(read: E, duration: Duration) -> Eventual<E::Output>
@@ -86,7 +181,7 @@ where
     Eventual::spawn(move |mut writer| async move {
         loop {
             let mut next = read.next().await?;
-            let end = time::Instant::now() + duration;
+            let end = tokio::time::Instant::now() + duration;
             loop {
                 // Allow replacing the value until the time is up. This
                 // necessarily introduces latency but de-duplicates when there
@@ -96,7 +191,7 @@ where
                     n = read.next() => {
                         next = n?;
                     }
-                    _ = time::sleep_until(end) => {
+                    _ = sleep_until(end) => {
                         break;
                     }
                 }
@@ -134,10 +229,11 @@ impl PipeHandle {
     }
 }
 
-pub fn handle_errors<E, F, Ok, Err>(source: E, f: F) -> Eventual<Ok>
+#[deprecated = "Not deterministic. This is a special case of filter. Retry should be better"]
+pub fn handle_errors<E, F, Ok, Err>(source: E, mut f: F) -> Eventual<Ok>
 where
     E: IntoReader<Output = Result<Ok, Err>>,
-    F: 'static + Send + Fn(Err),
+    F: 'static + Send + FnMut(Err),
     Ok: Value,
     Err: Value,
 {
@@ -181,16 +277,12 @@ where
             let mut next = e.next().await;
 
             loop {
-                match next {
-                    Ok(Ok(v)) => {
+                match next? {
+                    Ok(v) => {
                         writer.write(v);
                         next = e.next().await;
                     }
-                    // TODO: Is this how we want to handle closed?
-                    Err(e) => {
-                        return Err(e);
-                    }
-                    Ok(Err(err)) => {
+                    Err(err) => {
                         select! {
                             e_temp = f(Some(err)) => {
                                 e = e_temp.subscribe();
@@ -207,30 +299,29 @@ where
     })
 }
 
-pub fn map_with_retry<I, Ok, Err, F, Fut>(source: Eventual<I>, f: F) -> Eventual<Ok>
+pub fn map_with_retry<I, Ok, Err, F, Fut, E, FutE>(
+    source: Eventual<I>,
+    f: F,
+    on_err: E,
+) -> Eventual<Ok>
 where
     F: 'static + Clone + Send + Fn(I) -> Fut,
+    E: 'static + Clone + Send + Sync + Fn(Err) -> FutE,
     I: Value,
     Ok: Value,
     Err: Value,
     Fut: Send + Future<Output = Result<Ok, Err>>,
+    FutE: Send + Future<Output = ()>,
 {
     retry(move |e| {
         let reader = source.subscribe();
         let f = f.clone();
+        let on_err = on_err.clone();
         async move {
-            if e.is_some() {
-                // TODO: Configurable time via on_error, which will
-                // also allow eg: loggging.
-                tokio::time::sleep(Duration::from_secs(5)).await;
+            if let Some(e) = e {
+                on_err(e).await;
             }
             map(reader, f)
         }
     })
 }
-
-// TODO: Consider re-exporting ByAddress<Arc<T>>. One nice thing about
-// having a local version is that it would allow this lib to impl things
-// like Error if ByAddress isn't already.
-//
-// TODO: Eventual.value
