@@ -10,13 +10,17 @@ use std::{
 mod change;
 mod eventual;
 mod eventual_ext;
+mod ptr;
 mod reader;
 mod shared_state;
 mod writer;
-use change::Change;
 
 pub(self) use {crate::Value, shared_state::*};
 
 pub use {
-    eventual::Eventual, eventual_ext::EventualExt, reader::EventualReader, writer::EventualWriter,
+    eventual::Eventual,
+    eventual_ext::{EventualExt, TryEventualExt},
+    ptr::Ptr,
+    reader::EventualReader,
+    writer::EventualWriter,
 };
